@@ -1,5 +1,5 @@
 #! /usr/bin/python3
-# Copyright © 2010-2013 Piotr Ożarowski <piotr@debian.org>
+# Copyright © 2010-2015 Piotr Ożarowski <piotr@debian.org>
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -21,25 +21,43 @@
 
 import re
 import sys
+try:
+    from distro_info import DistroInfo  # python3-distro-info package
+except ImportError:
+    DistroInfo = None
 from gzip import decompress
 from os import chdir, mkdir
 from os.path import dirname, exists, isdir, join, split
 from urllib.request import urlopen
 
-SOURCE = 'http://ftp.debian.org/debian/dists/unstable/main/Contents-amd64.gz'
+if '--ubuntu' in sys.argv and DistroInfo:
+    SOURCES = [
+        'http://archive.ubuntu.com/ubuntu/dists/%s/Contents-amd64.gz' %
+        DistroInfo('ubuntu').devel(),
+    ]
+else:
+    SOURCES = [
+        'http://ftp.debian.org/debian/dists/unstable/main/Contents-all.gz',
+        'http://ftp.debian.org/debian/dists/unstable/main/Contents-amd64.gz',
+    ]
+
 IGNORED_PKGS = {'python-setuptools', 'python3-setuptools', 'pypy-setuptools'}
-DEFAULTS = {
-    'cpython2': [
-        'python python\n',
-        'setuptools python-pkg-resources\n',
-        'wsgiref python (>= 2.5) | python-wsgiref\n',
-        'argparse python (>= 2.7) | python-argparse\n',
+OVERRIDES = {
+    'cpython2': {
+        'python': 'python',
+        'setuptools': 'python-pkg-resources',
+        'wsgiref': 'python (>= 2.5) | python-wsgiref',
+        'argparse': 'python (>= 2.7) | python-argparse',
         # not recognized due to .pth file (egg-info is in PIL/ and not in *-packages/)
-        'pil python-imaging\n'],
-    'cpython3': [
-        'setuptools python3-pkg-resources\n',
-        'argparse python3 (>= 3.2)\n'],
-    'pypy': []
+        'pil': 'python-pil',
+        'Pillow': 'python-pil'},
+    'cpython3': {
+        'pil': 'python3-pil',
+        'Pillow': 'python3-pil',
+        'pylint': 'pylint',
+        'setuptools': 'python3-pkg-resources',
+        'argparse': 'python3 (>= 3.2)'},
+    'pypy': {}
 }
 
 public_egg = re.compile(r'''
@@ -56,7 +74,7 @@ public_egg = re.compile(r'''
             (lib/pypy/dist-packages)
         )
     )
-    /[^/]*\.egg-info
+    /[^/]*\.(dist|egg)-info
 ''', re.VERBOSE).match
 
 skip_sensible_names = True if '--skip-sensible-names' in sys.argv else False
@@ -68,29 +86,40 @@ else:
     sys.path.append('/usr/share/dh-python/dhpython/')
 from dhpython.pydist import sensible_pname
 
+data = ''
 if not isdir('cache'):
     mkdir('cache')
-cache_fpath = join('cache', split(SOURCE)[-1])
-if not exists(cache_fpath):
-    data = urlopen(SOURCE).read()
-    with open(cache_fpath, 'wb') as fp:
-        fp.write(data)
-else:
-    data = open(cache_fpath, 'rb').read()
-data = str(decompress(data), encoding='UTF-8')
+for source in SOURCES:
+    cache_fpath = join('cache', split(source)[-1])
+    if not exists(cache_fpath):
+        with urlopen(source) as fp:
+            source_data = fp.read()
+        with open(cache_fpath, 'wb') as fp:
+            fp.write(source_data)
+    else:
+        with open(cache_fpath, 'rb') as fp:
+            source_data = fp.read()
+    try:
+        data += str(decompress(source_data), encoding='UTF-8')
+    except UnicodeDecodeError as e:  # Ubuntu
+        data += str(decompress(source_data), encoding='ISO-8859-15')
 
 result = {
-    'cpython2': {},
-    'cpython3': {},
-    'pypy': {}}
+    'cpython3': {}
+}
 
-is_header = True
+# Contents file doesn't contain comment these days
+is_header = not data.startswith('bin')
 for line in data.splitlines():
     if is_header:
         if line.startswith('FILE'):
             is_header = False
         continue
-    path, desc = line.rsplit(maxsplit=1)
+    try:
+        path, desc = line.rsplit(maxsplit=1)
+    except ValueError:
+        # NOTE(jamespage) some lines in Ubuntu are not parseable.
+        continue
     path = '/' + path.rstrip()
     section, pkg_name = desc.rsplit('/', 1)
     if pkg_name in IGNORED_PKGS:
@@ -98,7 +127,7 @@ for line in data.splitlines():
     match = public_egg(path)
     if match:
         egg_name = [i.split('-', 1)[0] for i in path.split('/')
-                    if i.endswith('.egg-info')][0]
+                    if i.endswith(('.egg-info', '.dist-info'))][0]
         if egg_name.endswith('.egg'):
             egg_name = egg_name[:-4]
 
@@ -114,8 +143,11 @@ for line in data.splitlines():
 
 for impl, details in result.items():
     with open('{}_fallback'.format(impl), 'w') as fp:
-        result = DEFAULTS[impl]
-        if result:
-            fp.writelines(result)
-        result = sorted('{} {}\n'.format(egg, pkg) for egg, pkg in details.items())
-        fp.writelines(result)
+        overrides = OVERRIDES[impl]
+        lines = []
+        for egg, value in overrides.items():
+            lines.append('{} {}\n'.format(egg, value))
+        lines.extend(
+            '{} {}\n'.format(egg, pkg) for egg, pkg in details.items() if egg not in overrides
+        )
+        fp.writelines(sorted(lines))
